@@ -887,6 +887,7 @@ extern Char **VG_(env_setenv)   ( Char ***envp, const Char* varname,
                                   const Char *val );
 extern void   VG_(env_unsetenv) ( Char **env, const Char *varname );
 extern void   VG_(env_remove_valgrind_env_stuff) ( Char** env ); 
+extern Char **VG_(env_clone)    ( Char **envp );
 
 extern void   VG_(nanosleep)(struct vki_timespec *);
 /* ---------------------------------------------------------------------
@@ -1118,6 +1119,7 @@ extern Addr VG_(client_end);
 extern Addr VG_(client_mapbase); /* base of mappings */
 extern Addr VG_(clstk_base);	/* client stack range */
 extern Addr VG_(clstk_end);
+extern UWord VG_(clstk_id);     /* client stack id */
 extern Addr VG_(client_trampoline_code);
 
 extern Addr VG_(brk_base);	/* start of brk */
@@ -1244,6 +1246,10 @@ extern Segment *VG_(split_segment)(Addr a);
 extern void VG_(pad_address_space)  (Addr start);
 extern void VG_(unpad_address_space)(Addr start);
 
+extern UWord VG_(handle_stack_register)(Addr start, Addr end);
+extern void VG_(handle_stack_deregister)(UWord id);
+extern void VG_(handle_stack_change)(UWord id, Addr start, Addr end);
+
 extern REGPARM(1)
    void VG_(unknown_SP_update) ( Addr new_SP );
 
