@@ -1027,7 +1027,7 @@ const struct SyscallTableEntry VGA_(syscall_table)[] = {
    GENX_(__NR_utimes,           sys_utimes),          // 251
    GENXY(__NR_statfs64,         sys_statfs64),        // 252
    GENXY(__NR_fstatfs64,        sys_fstatfs64),       // 253
-   //   (__NR_fadvise64_64,     sys_fadvise64_64),    // 254
+   LINX_(__NR_fadvise64_64,     sys_fadvise64_64),    // 254
 
    //   (__NR_rtas,		sys_ni_syscall),      // 255 ppc/Linux (64b)
    //   (__NR_debug_setcontext, sys_ni_syscall),      // 256 ppc/Linux
@@ -1035,8 +1035,8 @@ const struct SyscallTableEntry VGA_(syscall_table)[] = {
    //   (__NR_remap_file_pages,	sys_ni_syscall),      // 258
    //   (__NR_mbind,            sys_mbind),           // 259
 
-   //   (__NR_get_mempolicy,    sys_get_mempolicy),   // 260
-   //   (__NR_set_mempolicy,    sys_set_mempolicy),   // 261
+   LINXY(__NR_get_mempolicy,    sys_get_mempolicy),   // 260
+   LINX_(__NR_set_mempolicy,    sys_set_mempolicy),   // 261
    GENXY(__NR_mq_open,          sys_mq_open),         // 262
    GENX_(__NR_mq_unlink,        sys_mq_unlink),       // (mq_open+1)
    GENX_(__NR_mq_timedsend,     sys_mq_timedsend),    // (mq_open+2)
@@ -1045,6 +1045,17 @@ const struct SyscallTableEntry VGA_(syscall_table)[] = {
    GENX_(__NR_mq_notify,        sys_mq_notify),       // (mq_open+4)
    GENXY(__NR_mq_getsetattr,    sys_mq_getsetattr),   // (mq_open+5)
    //   (__NR_sys_kexec_load,   sys_ni_syscall),      // 268
+//   LINX_(__NR_add_key,	sys_add_key),	      // 269
+
+//   LINX_(__NR_request_key,	sys_request_key),     // 270
+//   LINXY(__NR_keyctl,		sys_keyctl),	      // 271
+// GENXY(__NR_waitid,		sys_waitid),	      // 272
+// LINX_(__NR_ioprio_set,	sys_ioprio_set),      // 273
+// LINX_(__NR_ioprio_get,	sys_ioprio_get),      // 274
+
+   LINX_(__NR_inotify_init,	sys_inotify_init),    // 275
+   LINX_(__NR_inotify_add_watch,sys_inotify_add_watch), // 276
+   LINX_(__NR_inotify_rm_watch,	sys_inotify_rm_watch), // 277
 };
 
 const UInt VGA_(syscall_table_size) = 
