@@ -477,7 +477,7 @@ static void emit_fpuop(UInt instr)
     subop = (instr >> 1) & 0x3ff;	/* X-form minor opcode */
 
     /* load up FPSCR if the instruction uses or sets it */
-    if (op != 8 && op != 23) {
+    if (op != 8 && op != 23 && !fpscr_dirty) {
 	emit(DFORM(LFD, 0, TST, TSTOFF(m_fpr) + 256));
 	emit(XFORM63(0x0f, 0x1e, 0, MO_MTFSF, 0));
 	if (dis)
@@ -884,6 +884,16 @@ static void emit_uinstr(UCodeBlock *cb, int i)
 	    VG_(printf)("\tcntlzw r%d,r%d\n", dreg, sreg);
 	break;
 
+    case RDTB:
+	vg_assert(u->tag1 == Lit16);
+	vg_assert(u->tag2 == RealReg);
+	sreg = u->val1;
+	dreg = VG_(real_dest_reg)(u->val2);
+	emit(XFORM31(dreg, sreg & 0x1f, (sreg >> 5) & 0x1f, MO_MFTB, 0));
+	if (dis)
+	    VG_(printf)("\tmftb%s r%d\n", (sreg & 1)? "u": "", dreg);
+	break;
+
     case AND:
 	vg_assert(u->tag2 == RealReg);
 	treg = VG_(real_source_reg)(u->val2);
