@@ -511,7 +511,13 @@ static int dis_opc_31(UCodeBlock *cb, int dis, unsigned int instr)
 	break;
 
     case 0x173:		/* mftb/mftbu */
-	goto bad;
+	reg = (rb << 5) + ra;
+	if (reg != 268 && reg != 269)
+	    goto bad;
+	t1 = new_temp(cb);
+	uInstr2(cb, RDTB, 4, Lit16, reg,  TempReg, t1);
+	uInstr2(cb, PUT,  4, TempReg, t1, ArchReg, rd);
+	break;
 
     case 0x153:		/* mfspr */
     case 0x1d3:		/* mtspr */
@@ -1194,6 +1200,9 @@ static void dis_instr(UCodeBlock *cb, Addr nip, int dis, Bool *isend)
 	    if (instr & 1)
 		set_lr(cb, nip + 4);
 	    dis_condition(cb, instr, nip + 4);
+	    /* bclr always aligns the dest PC, even if LR is not aligned */
+	    uInstr2(cb, AND,   4, Literal, 0, TempReg, t1);
+	    uLiteral(cb, ~0x3UL);
 	    uInstr1(cb, JMP,   0, TempReg, t1);
 	    if (rd & 0x10) {
 		/* unconditional blr[l] or bd[n]zlr[l] */
@@ -1231,6 +1240,9 @@ static void dis_instr(UCodeBlock *cb, Addr nip, int dis, Bool *isend)
 	    t1 = new_temp(cb);
 	    uInstr2(cb, GET,   4, ArchReg, R_CTR, TempReg, t1);
 	    dis_condition(cb, instr, nip + 4);
+	    /* bctr always aligns the dest PC, even if CTR is not aligned */
+	    uInstr2(cb, AND,   4, Literal, 0, TempReg, t1);
+	    uLiteral(cb, ~0x3UL);
 	    uInstr1(cb, JMP,   0, TempReg, t1);
 	    if (rd & 0x10) {
 		/* unconditional bctr[l] or bd[n]zctr[l] */
