@@ -95,6 +95,7 @@ Addr VG_(client_mapbase);
 Addr VG_(client_trampoline_code);
 Addr VG_(clstk_base);
 Addr VG_(clstk_end);
+UWord VG_(clstk_id);             /* client stack id */
 
 Addr VG_(brk_base);	         /* start of brk */
 Addr VG_(brk_limit);	         /* current brk */
@@ -2699,6 +2700,11 @@ int main(int argc, char **argv, char **envp)
    if (VG_(clo_pointercheck))
       VG_(clo_pointercheck) = VGA_(setup_pointercheck)();
 
+   //--------------------------------------------------------------
+   // register client stack
+   //--------------------------------------------------------------
+   VG_(clstk_id) = VG_(handle_stack_register)(VG_(clstk_base), VG_(clstk_end));
+
    //--------------------------------------------------------------
    // Run!
    //--------------------------------------------------------------
