@@ -761,26 +761,23 @@ Long VG_(atoll16) ( Char* str )
    return n;
 }
 
-Long VG_(atoll36) ( UInt base, Char* str )
+Long VG_(atoll36) ( Char* str )
 {
    Bool neg = False;
    Long n = 0;
-   vg_assert(base >= 2 && base <= 36);
    if (*str == '-') { str++; neg = True; };
    while (True) {
-      if (*str >= '0' 
-          && *str <= (Char)('9' - (10 - base))) {
-         n = base*n + (Long)(*str - '0');
+      Char c = *str;
+      if (c >= '0' && c <= (Char)'9') {
+         n = 36*n + (Long)(c - '0');
       }
       else 
-      if (base > 10 && *str >= 'A' 
-          && *str <= (Char)('Z' - (36 - base))) {
-         n = base*n + (Long)((*str - 'A') + 10);
+      if (c >= 'A' && c <= (Char)'Z') {
+         n = 36*n + (Long)((c - 'A') + 10);
       }
       else 
-      if (base > 10 && *str >= 'a' 
-          && *str <= (Char)('z' - (36 - base))) {
-         n = base*n + (Long)((*str - 'a') + 10);
+      if (c >= 'a' && c <= (Char)'z') {
+         n = 36*n + (Long)((c - 'a') + 10);
       }
       else {
 	break;
@@ -792,6 +789,7 @@ Long VG_(atoll36) ( UInt base, Char* str )
 }
 
 
+
 Char* VG_(strcat) ( Char* dest, const Char* src )
 {
    Char* dest_orig = dest;
@@ -1359,7 +1357,7 @@ Bool VG_(getcwd_alloc) ( Char** out )
    ------------------------------------------------------------------ */
 
 /* clone the environment */
-static Char **env_clone ( Char **oldenv )
+Char **VG_(env_clone) ( Char **oldenv )
 {
    Char **oldenvp;
    Char **newenvp;
@@ -1378,8 +1376,8 @@ static Char **env_clone ( Char **oldenv )
    while (oldenvp && *oldenvp) {
       *newenvp++ = *oldenvp++;
    }
-   
-   *newenvp = *oldenvp;
+
+   *newenvp = NULL;
 
    return newenv;
 }
@@ -1396,7 +1394,7 @@ void  VG_(env_unsetenv) ( Char **env, const Char *varname )
 	 to++;
       }
    }
-   *to = *from;
+   *to = NULL;
 }
 
 /* set the environment; returns the old env if a new one was allocated */
@@ -1643,7 +1641,7 @@ Int VG_(system) ( Char* cmd )
       /* restore the DATA rlimit for the child */
       VG_(setrlimit)(VKI_RLIMIT_DATA, &VG_(client_rlimit_data));
 
-      envp = env_clone(VG_(client_envp));
+      envp = VG_(env_clone)(VG_(client_envp));
       VG_(env_remove_valgrind_env_stuff)( envp ); 
 
       argv[0] = "/bin/sh";
