@@ -236,7 +236,7 @@ ThreadId VG_(get_lwp_tid)(Int lwp)
 {
    ThreadId tid;
    
-   for(tid = 1; tid <= VG_N_THREADS; tid++)
+   for(tid = 1; tid < VG_N_THREADS; tid++)
       if (VG_(threads)[tid].status != VgTs_Empty && VG_(threads)[tid].os_state.lwpid == lwp)
 	 return tid;
 
@@ -946,6 +946,21 @@ void do_client_request ( ThreadId tid )
             SET_CLREQ_RETVAL( tid, count );
          break; }
 
+      case VG_USERREQ__STACK_REGISTER: {
+         UWord sid = VG_(handle_stack_register)((Addr)arg[1], (Addr)arg[2]);
+         SET_CLREQ_RETVAL( tid, sid );
+         break; }
+
+      case VG_USERREQ__STACK_DEREGISTER: {
+         VG_(handle_stack_deregister)(arg[1]);
+         SET_CLREQ_RETVAL( tid, 0 );     /* return value is meaningless */
+         break; }
+
+      case VG_USERREQ__STACK_CHANGE: {
+         VG_(handle_stack_change)(arg[1], (Addr)arg[2], (Addr)arg[3]);
+         SET_CLREQ_RETVAL( tid, 0 );     /* return value is meaningless */
+         break; }
+
       case VG_USERREQ__GET_MALLOCFUNCS: {
 	 struct vg_mallocfunc_info *info = (struct vg_mallocfunc_info *)arg[1];
 
