@@ -712,7 +712,16 @@ void VG_(do_sys_sigprocmask) ( ThreadId tid,
       vg_assert(VG_(is_valid_tid)(tid));
       do_setmask ( tid, how, set, oldset );
       SET_SYSCALL_RETVAL(tid, 0);
+#ifndef __powerpc__
+      /* On PPC, gpr3 is used to hold both the first argument to a function and
+       * the syscall return value.  If we poll for signals here and find one,
+       * we will put the first arg (sigNo) into gpr3.  But later, before the
+       * signal handler gets run, we overwrite that with the sys_sigprocmask
+       * return value.  So on PPC, we can't poll for signals here; we wait
+       * until POST(sys_sigprocmask) instead. */
+
       VG_(poll_signals)(tid);	/* look for any newly deliverable signals */
+#endif
       break;
 
    default:
@@ -1699,6 +1708,10 @@ Bool VG_(extend_stack)(Addr addr, UInt maxsize)
 		 -1, 0) == (void *)-1)
       return False;
 
+   /* When we change the main stack, we have to let the stack handling
+      code know about it. */
+   VG_(handle_stack_change)(VG_(clstk_id), base, VG_(clstk_end));
+
    if (0)
       VG_(printf)("extended stack: %p %d\n",
 		  base, newsize);
