@@ -7,7 +7,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -28,13 +28,14 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
+#include "core.h"
 #include "vg_symtypes.h"
 #include "vg_symtab2.h"
-#include "ume_archdefs.h"	/* for ELF_ENDIANNESS etc. */
 
 #include <elf.h>          /* ELF defns                      */
 
+static SegInfo* segInfo = NULL;
+
 static Bool
 intercept_demangle(const Char*, Char*, Int);
 
@@ -216,10 +217,12 @@ void VG_(addLineInfo) ( SegInfo* si,
     * correspond to one line, unlike with C-level debug info where
     * multiple instructions can map to the one line), but avoid
     * catching any other instructions bogusly. */
-   if (this > next && VG_(clo_verbosity) > 2) {
-       VG_(message)(Vg_DebugMsg, 
-                    "warning: line info addresses out of order "
-                    "at entry %d: 0x%x 0x%x", entry, this, next);
+   if (this > next) {
+       if (VG_(clo_verbosity) > 2) {
+           VG_(message)(Vg_DebugMsg, 
+                        "warning: line info addresses out of order "
+                        "at entry %d: 0x%x 0x%x", entry, this, next);
+       }
        size = 1;
    }
 
@@ -405,72 +408,48 @@ static Int compare_RiSym(void *va, void *vb) {
 
 /* Two symbols have the same address.  Which name do we prefer?
 
-   In general we prefer the longer name, but if the choice is between
-   __libc_X and X, then choose X (similarly with __GI__ and __
-   prefixes).
+   The general rule is to prefer the shorter symbol name.  If the
+   symbol contains a '@', which means its versioned, then the length
+   up to the '@' is used for length comparison purposes (so
+   "foo@GLIBC_2.4.2" is considered shorter than "foobar"), but if two
+   symbols have the same length, the one with the version string is
+   preferred.  If all else fails, use alphabetical ordering.
  */
 static RiSym *prefersym(RiSym *a, RiSym *b)
 {
-   Int pfx;
-   Int lena, lenb;
-   Int i;
-   static const struct {
-      const Char *prefix;
-      Int len;
-   } prefixes[] = {
-#define PFX(x)	{ x, sizeof(x)-1 }
-      /* order from longest to shortest */
-      PFX("__GI___libc_"),
-      PFX("__GI___"),
-      PFX("__libc_"),
-      PFX("__GI__"),
-      PFX("__GI_"),
-      PFX("__"),
-#undef PFX
-   };
+   Int lena, lenb;		/* full length */
+   Int vlena, vlenb;		/* length without version */
+   const Char *vpa, *vpb;
 
-   lena = VG_(strlen)(a->name);
-   lenb = VG_(strlen)(b->name);
+   vlena = lena = VG_(strlen)(a->name);
+   vlenb = lenb = VG_(strlen)(b->name);
 
-   /* rearrange so that a is the long one */
-   if (lena < lenb) {
-      RiSym *t;
-      Int lt;
+   vpa = VG_(strchr)(a->name, '@');
+   vpb = VG_(strchr)(b->name, '@');
 
-      t = a;
-      a = b;
-      b = t;
+   if (vpa)
+      vlena = vpa - a->name;
+   if (vpb)
+      vlenb = vpb - b->name;
 
-      lt = lena;
-      lena = lenb;
-      lenb = lt;
-   }
-
-   /* Ugh.  If we get a "free", always choose it.  This is because
-      normally, this function would choose "cfree" over free.  cfree is
-      an alias for free.  If there's any more symbols like this, we may
-      want to consider making this mechanism more generic.
-    */
-   if(VG_(strcmp)(a->name, "free") == 0)
-       return a;
-
-   if(VG_(strcmp)(b->name, "free") == 0)
-       return b;
-
-   for(i = pfx = 0; i < sizeof(prefixes)/sizeof(*prefixes); i++) {
-      Int pfxlen = prefixes[i].len;
-
-      if (pfxlen < lena &&
-	  VG_(memcmp)(a->name, prefixes[i].prefix, pfxlen) == 0) {
-	 pfx = pfxlen;
-	 break;
-      }
-   }
-
-   if (pfx != 0 && VG_(strcmp)(a->name + pfx, b->name) == 0)
+   /* Select the shortest unversioned name */
+   if (vlena < vlenb)
+      return a;
+   else if (vlenb < vlena)
       return b;
 
-   return a;
+   /* Equal lengths; select the versioned name */
+   if (vpa && !vpb)
+      return a;
+   if (vpb && !vpa)
+      return b;
+
+   /* Either both versioned or neither is versioned; select them
+      alphabetically */
+   if (VG_(strcmp)(a->name, b->name) < 0)
+      return a;
+   else
+      return b;
 }
 
 static 
@@ -491,11 +470,12 @@ void canonicaliseSymtab ( SegInfo* si )
       if(VG_(strncmp)(si->symtab[i].name, VG_INTERCEPT_PREFIX,
                       VG_INTERCEPT_PREFIX_LEN) == 0) {
          int len = VG_(strlen)(si->symtab[i].name);
-         char *buf = VG_(malloc)(len), *colon;
+         char *buf = VG_(arena_malloc)(VG_AR_SYMTAB, len), *colon;
          intercept_demangle(si->symtab[i].name, buf, len);
 	 colon = buf + VG_(strlen)(buf) - 1;
 	 while(*colon != ':') colon--;
 	 VG_(strncpy_safely)(si->symtab[i].name, colon+1, len);
+	 VG_(arena_free)(VG_AR_SYMTAB, buf);
       }
    }
 
@@ -759,11 +739,11 @@ Bool VG_(is_object_file)(const void *buf)
              && ehdr->e_ident[EI_MAG1] == 'E'
              && ehdr->e_ident[EI_MAG2] == 'L'
              && ehdr->e_ident[EI_MAG3] == 'F');
-      ok &= (ehdr->e_ident[EI_CLASS] == ELF_CLASS
-             && ehdr->e_ident[EI_DATA] == ELF_ENDIANNESS
+      ok &= (ehdr->e_ident[EI_CLASS] == VG_ELF_CLASS
+             && ehdr->e_ident[EI_DATA] == VG_ELF_ENDIANNESS
              && ehdr->e_ident[EI_VERSION] == EV_CURRENT);
       ok &= (ehdr->e_type == ET_EXEC || ehdr->e_type == ET_DYN);
-      ok &= (ehdr->e_machine == ELF_MACHINE);
+      ok &= (ehdr->e_machine == VG_ELF_MACHINE);
       ok &= (ehdr->e_version == EV_CURRENT);
       ok &= (ehdr->e_shstrndx != SHN_UNDEF);
       ok &= (ehdr->e_shoff != 0 && ehdr->e_shnum != 0);
@@ -831,15 +811,11 @@ intercept_demangle(const Char* symbol, Char* result, Int nbytes)
    return True;
 }
 
-// Forward declaration
-static void add_redirect_addr(const Char *from_lib, const Char *from_sym,
-			      Addr to_addr);
-
 static
 void handle_intercept( SegInfo* si, Char* symbol, Elf32_Sym* sym)
 {
    int len = VG_(strlen)(symbol) + 1 - VG_INTERCEPT_PREFIX_LEN;
-   char *lib = VG_(malloc)(len);
+   char *lib = VG_(arena_malloc)(VG_AR_SYMTAB, len);
    Char *func;
 
    intercept_demangle(symbol, lib, len);
@@ -848,14 +824,28 @@ void handle_intercept( SegInfo* si, Char* symbol, Elf32_Sym* sym)
    while(*func != ':') func--;
    *func = '\0';
 
-   add_redirect_addr(lib, func+1, si->offset + sym->st_value);
-   VG_(free)(lib);
+   VG_(add_redirect_addr)(lib, func+1, si->offset + sym->st_value);
+   VG_(arena_free)(VG_AR_SYMTAB, lib);
+}
+
+Bool VG_(resolve_redir_allsegs)(CodeRedirect *redir)
+{
+   SegInfo *si;
+
+   for(si = segInfo; si != NULL; si = si->next)
+      if (VG_(resolve_redir)(redir, si))
+	 return True;
+
+   return False;
 }
 
 static
 void handle_wrapper( SegInfo* si, Char* symbol, Elf32_Sym* sym)
 {
-   VG_(intercept_libc_freeres_wrapper)((Addr)(si->offset + sym->st_value));
+   if (VG_(strcmp)(symbol, STR(VG_WRAPPER(freeres))) == 0)
+      VGA_(intercept_libc_freeres_wrapper)((Addr)(si->offset + sym->st_value));
+   else if (VG_(strcmp)(symbol, STR(VG_WRAPPER(pthread_startfunc_wrapper))) == 0)
+      VG_(pthread_startfunc_wrapper)((Addr)(si->offset + sym->st_value));
 }
 
 /* Read a symbol table (normal or dynamic) */
@@ -1094,6 +1084,7 @@ Addr open_debug_file( Char* name, UInt crc, UInt* size )
    Int fd;
    struct vki_stat stat_buf;
    Addr addr;
+   UInt calccrc;
 
    if ((fd = VG_(open)(name, VKI_O_RDONLY, 0)) < 0)
       return 0;
@@ -1103,6 +1094,9 @@ Addr open_debug_file( Char* name, UInt crc, UInt* size )
       return 0;
    }
 
+   if (VG_(clo_verbosity) > 1)
+      VG_(message)(Vg_UserMsg, "Reading debug info from %s...", name);
+
    *size = stat_buf.st_size;
    
    if ((addr = (Addr)VG_(mmap)(NULL, *size, VKI_PROT_READ,
@@ -1115,9 +1109,12 @@ Addr open_debug_file( Char* name, UInt crc, UInt* size )
 
    VG_(close)(fd);
    
-   if (calc_gnu_debuglink_crc32(0, (UChar*)addr, *size) != crc) {
+   calccrc = calc_gnu_debuglink_crc32(0, (UChar*)addr, *size);
+   if (calccrc != crc) {
       int res = VG_(munmap)((void*)addr, *size);
       vg_assert(0 == res);
+      if (VG_(clo_verbosity) > 1)
+	 VG_(message)(Vg_UserMsg, "... CRC mismatch (computed %08x wanted %08x)", calccrc, crc);
       return 0;
    }
    
@@ -1130,7 +1127,7 @@ Addr open_debug_file( Char* name, UInt crc, UInt* size )
 static
 Addr find_debug_file( Char* objpath, Char* debugname, UInt crc, UInt* size )
 {
-   Char *objdir = VG_(strdup)(objpath);
+   Char *objdir = VG_(arena_strdup)(VG_AR_SYMTAB, objpath);
    Char *objdirptr;
    Char *debugpath;
    Addr addr = 0;
@@ -1138,7 +1135,7 @@ Addr find_debug_file( Char* objpath, Char* debugname, UInt crc, UInt* size )
    if ((objdirptr = VG_(strrchr)(objdir, '/')) != NULL)
       *objdirptr = '\0';
 
-   debugpath = VG_(malloc)(VG_(strlen)(objdir) + VG_(strlen)(debugname) + 16);
+   debugpath = VG_(arena_malloc)(VG_AR_SYMTAB, VG_(strlen)(objdir) + VG_(strlen)(debugname) + 16);
    
    VG_(sprintf)(debugpath, "%s/%s", objdir, debugname);
 
@@ -1150,8 +1147,8 @@ Addr find_debug_file( Char* objpath, Char* debugname, UInt crc, UInt* size )
       }
    }
 
-   VG_(free)(debugpath);
-   VG_(free)(objdir);
+   VG_(arena_free)(VG_AR_SYMTAB, debugpath);
+   VG_(arena_free)(VG_AR_SYMTAB, objdir);
    
    return addr;
 }
@@ -1298,8 +1295,8 @@ Bool vg_read_lib_symbols ( SegInfo* si )
 	       si->bss_size = 0;
 	 }
 
-	 mapped = mapped & ~(VKI_BYTES_PER_PAGE-1);
-	 mapped_end = (mapped_end + VKI_BYTES_PER_PAGE - 1) & ~(VKI_BYTES_PER_PAGE-1);
+	 mapped = mapped & ~(VKI_PAGE_SIZE-1);
+	 mapped_end = (mapped_end + VKI_PAGE_SIZE - 1) & ~(VKI_PAGE_SIZE-1);
 
 	 if (VG_(needs).data_syms &&
 	     (mapped >= si->start && mapped <= (si->start+si->size)) &&
@@ -1313,14 +1310,14 @@ Bool vg_read_lib_symbols ( SegInfo* si )
 			      si->start, si->start+si->size, si->size,
 			      si->start+newsz, newsz);
 
-	       for(seg = VG_(find_segment)(si->start);
+	       for(seg = VG_(find_segment_containing)(si->start);
 		   seg != NULL && VG_(seg_overlaps)(seg, si->start, si->size); 
 		   seg = VG_(next_segment)(seg)) {
 		  if (seg->symtab == si)
 		     continue;
 
 		  if (seg->symtab != NULL)
-		     VG_(symtab_decref)(seg->symtab, seg->addr, seg->len);
+		     VG_(symtab_decref)(seg->symtab, seg->addr);
 
 		  VG_(symtab_incref)(si);
 		  seg->symtab = si;
@@ -1431,7 +1428,7 @@ Bool vg_read_lib_symbols ( SegInfo* si )
 
       /* Did we find a debuglink section? */
       if (debuglink != NULL) {
-         UInt crc_offset = (VG_(strlen)(debuglink) + 4) & ~3;
+         UInt crc_offset = ROUNDUP(VG_(strlen)(debuglink)+1, 4);
          UInt crc;
 
          vg_assert(crc_offset + sizeof(UInt) <= debuglink_sz);
@@ -1527,10 +1524,6 @@ Bool vg_read_lib_symbols ( SegInfo* si )
    address ranges, and as a result the SegInfos in this list describe
    disjoint address ranges. 
 */
-static SegInfo* segInfo = NULL;
-
-static void resolve_seg_redirs(SegInfo *si);
-
 SegInfo *VG_(read_seg_symbols) ( Segment *seg )
 {
    SegInfo* si;
@@ -1586,7 +1579,7 @@ SegInfo *VG_(read_seg_symbols) ( Segment *seg )
       canonicaliseScopetab ( si );
 
       /* do redirects */
-      resolve_seg_redirs( si );
+      VG_(resolve_seg_redirs)( si );
    }
    VGP_POPCC(VgpReadSyms);
 
@@ -1600,7 +1593,7 @@ SegInfo *VG_(read_seg_symbols) ( Segment *seg )
    accuracy of error messages, but we need to do it in order to
    maintain the no-overlapping invariant.
 */
-static void unload_symbols ( Addr start, UInt length )
+static void unload_symbols ( Addr start, SizeT length )
 {
    SegInfo *prev, *curr;
 
@@ -1634,7 +1627,7 @@ static void unload_symbols ( Addr start, UInt length )
    return;
 }
 
-void VG_(symtab_decref)(SegInfo *si, Addr start, UInt len)
+void VG_(symtab_decref)(SegInfo *si, Addr start)
 {
    vg_assert(si->ref >= 1);
    if (--si->ref == 0)
@@ -1685,8 +1678,7 @@ static Int search_one_symtab ( SegInfo* si, Addr ptr,
    table is designed we have no option but to do a complete linear
    scan of the table.  Returns NULL if not found. */
 
-static Addr reverse_search_one_symtab ( const SegInfo* si,
-                                        const Char* name )
+Addr VG_(reverse_search_one_symtab) ( const SegInfo* si, const Char* name )
 {
    UInt i;
    for (i = 0; i < si->symtab_used; i++) {
@@ -1713,9 +1705,9 @@ static void search_all_symtabs ( Addr ptr, /*OUT*/SegInfo** psi,
 
    VGP_PUSHCC(VgpSearchSyms);
 
-   s = VG_(find_segment)(ptr);
+   s = VG_(find_segment_containing)(ptr);
 
-   if (s == NULL || !VG_(seg_overlaps)(s, ptr, 0) || s->symtab == NULL)
+   if (s == NULL || s->symtab == NULL)
       goto not_found;
    
    si = s->symtab;
@@ -2003,10 +1995,7 @@ static UInt *regaddr(ThreadId tid, Int regno)
 {
    UInt *ret = 0;
 
-   if (VG_(is_running_thread)(tid))
-      ret = VG_(arch_reg_addr_from_BB)(regno);
-   else
-      ret = VG_(arch_reg_addr)(regno, &VG_(threads)[tid].arch);
+   ret = VGA_(reg_addr_from_tst)(regno, &VG_(threads)[tid].arch);
 
    if (ret == 0) {
       Char file[100];
@@ -2103,7 +2092,8 @@ Variable *VG_(get_scope_variables)(ThreadId tid)
 
 	 case SyEBPrel:
 	 case SyESPrel:
-	    reg = *regaddr(tid, sym->kind == SyESPrel ? R_STACK_PTR : R_FRAME_PTR);
+	    reg = *regaddr(tid, sym->kind == SyESPrel ? 
+                                R_STACK_PTR : R_FRAME_PTR);
 	    if (debug)
 	       VG_(printf)("reg=%p+%d=%p\n", reg, sym->u.offset, reg+sym->u.offset);
 	    v->valuep = (Addr)(reg + sym->u.offset);
@@ -2145,9 +2135,9 @@ Char* VG_(describe_eip)(Addr eip, Char* buf, Int n_buf)
    UInt  lineno; 
    UChar ibuf[20];
    UInt  n = 0;
-   UChar buf_fn[M_VG_ERRTXT];
-   UChar buf_obj[M_VG_ERRTXT];
-   UChar buf_srcloc[M_VG_ERRTXT];
+   static UChar buf_fn[M_VG_ERRTXT];
+   static UChar buf_obj[M_VG_ERRTXT];
+   static UChar buf_srcloc[M_VG_ERRTXT];
    Bool  know_fnname  = VG_(get_fnname) (eip, buf_fn,  M_VG_ERRTXT);
    Bool  know_objname = VG_(get_objname)(eip, buf_obj, M_VG_ERRTXT);
    Bool  know_srcloc  = VG_(get_filename_linenum)(eip, buf_srcloc, M_VG_ERRTXT, 
@@ -2185,7 +2175,7 @@ Char* VG_(describe_eip)(Addr eip, Char* buf, Int n_buf)
 void VG_(mini_stack_dump) ( Addr eips[], UInt n_eips )
 {
    UInt  i;
-   UChar buf[M_VG_ERRTXT];
+   static UChar buf[M_VG_ERRTXT];
    Bool  main_done = False;
 
    Int stop_at = n_eips;
@@ -2197,8 +2187,8 @@ void VG_(mini_stack_dump) ( Addr eips[], UInt n_eips )
    i = 0;
    do {
       Addr eip = eips[i];
-      if (i > 0)
-	 eip -= INSTR_MIN_SIZE;            /* point to calling line */
+      if (i > 0) 
+         eip -= MIN_INSTR_SIZE;     // point to calling line
       VG_(describe_eip)(eip, buf, M_VG_ERRTXT);
 
       if ( ! VG_(clo_show_below_main)) {
@@ -2215,316 +2205,6 @@ void VG_(mini_stack_dump) ( Addr eips[], UInt n_eips )
 }
 
 
-/*------------------------------------------------------------*/
-/*--- General purpose redirection.                         ---*/
-/*------------------------------------------------------------*/
-
-/* Set to True for debug printing. */
-static const Bool verbose_redir = False;
-
-
-/* resolved redirections, indexed by from_addr */
-typedef struct _CodeRedirect {
-   const Char	*from_lib;	/* library qualifier pattern */
-   const Char	*from_sym;	/* symbol */
-   Addr		from_addr;	/* old addr */
-
-   const Char	*to_lib;	/* library qualifier pattern */
-   const Char	*to_sym;	/* symbol */
-   Addr		to_addr;	/* new addr */
-
-   struct _CodeRedirect *next;	/* next pointer on unresolved list */
-} CodeRedirect;
-
-static Int addrcmp(const void *ap, const void *bp)
-{
-   Addr a = *(Addr *)ap;
-   Addr b = *(Addr *)bp;
-   Int ret;
-
-   if (a == b)
-      ret = 0;
-   else
-      ret = (a < b) ? -1 : 1;
-
-   return ret;
-}
-
-static Char *straddr(void *p)
-{
-   static Char buf[16];
-
-   VG_(sprintf)(buf, "%p", *(Addr *)p);
-
-   return buf;
-}
-
-static SkipList sk_resolved_redir = SKIPLIST_INIT(CodeRedirect, from_addr, 
-						  addrcmp, straddr, VG_AR_SYMTAB);
-static CodeRedirect *unresolved_redir = NULL;
-
-static Bool match_lib(const Char *pattern, const SegInfo *si)
-{
-   /* pattern == NULL matches everything, otherwise use globbing
-
-      If the pattern starts with:
-	file:, then match filename
-	soname:, then match soname
-	something else, match filename
-   */
-   const Char *name = si->filename;
-
-   if (pattern == NULL)
-      return True;
-
-   if (VG_(strncmp)(pattern, "file:", 5) == 0) {
-      pattern += 5;
-      name = si->filename;
-   }
-   if (VG_(strncmp)(pattern, "soname:", 7) == 0) {
-      pattern += 7;
-      name = si->soname;
-   }
-
-   if (name == NULL)
-      return False;
-   
-   return VG_(string_match)(pattern, name);
-}
-
-/* Resolve a redir using si if possible, and add it to the resolved
-   list */
-static Bool resolve_redir(CodeRedirect *redir, const SegInfo *si)
-{
-   Bool resolved;
-
-   vg_assert(si != NULL);
-   vg_assert(si->seg != NULL);
-
-   /* no redirection from Valgrind segments */
-   if (si->seg->flags & SF_VALGRIND)
-      return False;
-
-   resolved = (redir->from_addr != 0) && (redir->to_addr != 0);
-
-   if (0 && verbose_redir)
-      VG_(printf)("   consider FROM binding %s:%s -> %s:%s in %s(%s)\n",
-		  redir->from_lib, redir->from_sym,
-		  redir->to_lib, redir->to_sym,
-		  si->filename, si->soname);
-
-   vg_assert(!resolved);
-
-   if (redir->from_addr == 0) {
-      vg_assert(redir->from_sym != NULL);
-
-      if (match_lib(redir->from_lib, si)) {
-	 redir->from_addr = reverse_search_one_symtab(si, redir->from_sym);
-	 if (verbose_redir && redir->from_addr != 0)
-	    VG_(printf)("   bind FROM: %p = %s:%s\n", 
-                        redir->from_addr,redir->from_lib, redir->from_sym );
-      }
-   }
-
-   if (redir->to_addr == 0) {
-      vg_assert(redir->to_sym != NULL);
-
-      if (match_lib(redir->to_lib, si)) {
-	 redir->to_addr = reverse_search_one_symtab(si, redir->to_sym);
-	 if (verbose_redir && redir->to_addr != 0)
-	    VG_(printf)("   bind   TO: %p = %s:%s\n", 
-                        redir->to_addr,redir->to_lib, redir->to_sym );
-
-      }
-   }
-
-   resolved = (redir->from_addr != 0) && (redir->to_addr != 0);
-
-   if (0 && verbose_redir)
-      VG_(printf)("resolve_redir: %s:%s from=%p %s:%s to=%p\n",
-		  redir->from_lib, redir->from_sym, redir->from_addr, 
-		  redir->to_lib, redir->to_sym, redir->to_addr);
-
-   if (resolved) {
-      if (VG_(clo_verbosity) > 2 || verbose_redir) {
-	 VG_(message)(Vg_DebugMsg, "  redir resolved (%s:%s=%p -> ",
-		      redir->from_lib, redir->from_sym, redir->from_addr);
-	 VG_(message)(Vg_DebugMsg, "                  %s:%s=%p)",
-		      redir->to_lib, redir->to_sym, redir->to_addr);
-      }
-      
-      if (VG_(search_transtab)(redir->from_addr) != 0) {
-	/* For some given (from, to) redir, the "from" function got
-           called before the .so containing "to" became available.  We
-           know this because there is already a translation for the
-           entry point of the original "from".  So the redirect will
-           never actually take effect unless that translation is
-           discarded.  
-
-           Note, we only really need to discard the first bb of the
-           old entry point, and so we avoid the problem of having to
-           figure out how big that bb was -- since it is at least 1
-           byte of original code, we can just pass 1 as the original
-           size to invalidate_translations() and it will indeed get
-           rid of the translation. 
-
-           Note, this is potentially expensive -- discarding
-           translations causes complete unchaining.  
-         */
-         if (VG_(clo_verbosity) > 2) {
-            VG_(message)(Vg_UserMsg,   
-	       "Discarding translation due to redirect of already called function" );
-            VG_(message)(Vg_UserMsg,
-                "   %s (%p -> %p)",
-                redir->from_sym, redir->from_addr, redir->to_addr );
-         }
-	 VG_(invalidate_translations)(redir->from_addr, 1, True);
-      }
-
-      VG_(SkipList_Insert)(&sk_resolved_redir, redir);
-   }
-
-   return resolved;
-}
-
-/* Go through the complete redir list, resolving as much as possible with this SegInfo.
-
-    This should be called when a new SegInfo symtab is loaded.
- */
-static void resolve_seg_redirs(SegInfo *si)
-{
-   CodeRedirect **prevp = &unresolved_redir;
-   CodeRedirect *redir, *next;
-
-   if (verbose_redir)
-      VG_(printf)("Considering redirs to/from %s(soname=%s)\n",
-                  si->filename, si->soname);
-
-   /* visit each unresolved redir - if it becomes resolved, then
-      remove it from the unresolved list */
-   for(redir = unresolved_redir; redir != NULL; redir = next) {
-      next = redir->next;
-
-      if (resolve_redir(redir, si)) {
-	 *prevp = next;
-	 redir->next = NULL;
-      } else
-	 prevp = &redir->next;
-   }
-}
-
-static Bool resolve_redir_allsegs(CodeRedirect *redir)
-{
-   SegInfo *si;
-
-   for(si = segInfo; si != NULL; si = si->next)
-      if (resolve_redir(redir, si))
-	 return True;
-
-   return False;
-}
-
-/* Redirect a lib/symbol reference to a function at lib/symbol */
-static void add_redirect_sym(const Char *from_lib, const Char *from_sym,
-			     const Char *to_lib, const Char *to_sym)
-{
-   CodeRedirect *redir = VG_(SkipNode_Alloc)(&sk_resolved_redir);
-
-   redir->from_lib = VG_(arena_strdup)(VG_AR_SYMTAB, from_lib);
-   redir->from_sym = VG_(arena_strdup)(VG_AR_SYMTAB, from_sym);
-   redir->from_addr = 0;
-
-   redir->to_lib = VG_(arena_strdup)(VG_AR_SYMTAB, to_lib);
-   redir->to_sym = VG_(arena_strdup)(VG_AR_SYMTAB, to_sym);
-   redir->to_addr = 0;
-
-   if (VG_(clo_verbosity) >= 2)
-      VG_(message)(Vg_UserMsg, 
-                   "REDIRECT %s(%s) to %s(%s)",
-                   from_lib, from_sym, to_lib, to_sym);
-
-   if (!resolve_redir_allsegs(redir)) {
-      /* can't resolve immediately; add to list */
-      redir->next = unresolved_redir;
-      unresolved_redir = redir;
-   }
-}
-
-/* Redirect a lib/symbol reference to a function at lib/symbol */
-static void add_redirect_addr(const Char *from_lib, const Char *from_sym,
-			      Addr to_addr)
-{
-   CodeRedirect *redir = VG_(SkipNode_Alloc)(&sk_resolved_redir);
-
-   redir->from_lib = VG_(arena_strdup)(VG_AR_SYMTAB, from_lib);
-   redir->from_sym = VG_(arena_strdup)(VG_AR_SYMTAB, from_sym);
-   redir->from_addr = 0;
-
-   redir->to_lib = NULL;
-   redir->to_sym = NULL;
-   redir->to_addr = to_addr;
-
-   if (!resolve_redir_allsegs(redir)) {
-      /* can't resolve immediately; add to list */
-      redir->next = unresolved_redir;
-      unresolved_redir = redir;
-   }
-}
-
-Addr VG_(code_redirect)(Addr a)
-{
-   CodeRedirect *r = VG_(SkipList_Find)(&sk_resolved_redir, &a);
-
-   if (r == NULL || r->from_addr != a)
-      return a;
-
-   vg_assert(r->to_addr != 0);
-
-   return r->to_addr;
-}
-
-void VG_(setup_code_redirect_table) ( void )
-{
-   static const struct {
-      const Char *from, *to;
-   } redirects[] = {
-      { "__GI___errno_location",	"__errno_location"	},
-      { "__errno_location",		"__errno_location"	},
-      { "__GI___h_errno_location",	"__h_errno_location"	},
-      { "__h_errno_location",		"__h_errno_location"	},
-      { "__GI___res_state",		"__res_state"		},
-      { "__res_state",			"__res_state"		},
-   };
-   Int i;
-
-   for(i = 0; i < sizeof(redirects)/sizeof(*redirects); i++) {
-      add_redirect_sym("soname:libc.so.6",	 redirects[i].from,
-                       "soname:libpthread.so.0", redirects[i].to);
-   }
-
-#ifdef __i386__
-   /* Redirect _dl_sysinfo_int80, which is glibc's default system call
-      routine, to the routine in our trampoline page so that the
-      special sysinfo unwind hack in vg_execontext.c will kick in.
-   */
-   add_redirect_addr("soname:ld-linux.so.2", "_dl_sysinfo_int80",
-                     VG_(client_trampoline_code)+VG_(tramp_syscall_offset));
-#endif
-   
-   /* Overenthusiastic use of PLT bypassing by the glibc people also
-      means we need to patch the following functions to our own
-      implementations of said, in mac_replace_strmem.c.
-    */
-   add_redirect_sym("soname:libc.so.6", "stpcpy",
-                    "*vgpreload_memcheck.so*", "stpcpy");
-   add_redirect_sym("soname:libc.so.6", "strnlen",
-                    "*vgpreload_memcheck.so*", "strnlen");
-   add_redirect_sym("soname:ld-linux.so.2", "stpcpy",
-                    "*vgpreload_memcheck.so*", "stpcpy");
-   add_redirect_sym("soname:ld-linux.so.2", "strchr",
-                    "*vgpreload_memcheck.so*", "strchr");
-}
-
 /*------------------------------------------------------------*/
 /*--- SegInfo accessor functions                           ---*/
 /*------------------------------------------------------------*/
@@ -2541,7 +2221,7 @@ Addr VG_(seg_start)(const SegInfo* seg)
    return seg->start;
 }
 
-UInt VG_(seg_size)(const SegInfo* seg)
+SizeT VG_(seg_size)(const SegInfo* seg)
 {
    return seg->size;
 }
@@ -2551,7 +2231,7 @@ const UChar* VG_(seg_filename)(const SegInfo* seg)
    return seg->filename;
 }
 
-UInt VG_(seg_sym_offset)(const SegInfo* seg)
+ULong VG_(seg_sym_offset)(const SegInfo* seg)
 {
    return seg->offset;
 }
