@@ -1047,6 +1047,7 @@ Char* VG_(name_UOpcode) ( Bool upper, Opcode opc )
       case DIV:		return "DIV";
       case UDIV:	return "UDIV";
       case CNTLZ:	return "CNTLZ";
+      case RDTB:	return "RDTB";
       case VEC_R:       return "VEC_R";
       case VEC_W:       return "VEC_W";
       case VEC:         return "VEC";
@@ -1359,6 +1360,11 @@ void pp_UInstrWorker ( Int instrNo, UInstr* u, Bool ppRegsLiveness )
          VG_(pp_UOperand)(u, 1, u->size, False);
          break;
 
+      case RDTB:
+	 VG_(printf)("0x%x, ", u->val1);
+	 VG_(pp_UOperand)(u, 2, 4, False);
+	 break;
+
       /* Print a "(s)" after args passed on stack */
       case CCALL:
          if (u->has_ret_val) {
@@ -1595,6 +1601,7 @@ Int VG_(get_reg_usage) ( UInstr* u, Tag tag, Int* regs, Bool* isWrites )
       case DIV:
       case UDIV: RD(1); RD(2); WR(2); break;
       case CNTLZ: RD(1); WR(1); break;
+      case RDTB: WR(2); break;
 
       case VEC_R: case VEC_W: RD(2); break;
       case VEC: break;
@@ -1714,6 +1721,7 @@ Int maybe_uinstrReadsArchReg ( UInstr* u )
       case DIV: case UDIV: case CNTLZ:
       case VEC: case VEC_R: case VEC_W:
       case VEC_FROMREG: case VEC_TOREG:
+      case RDTB:
 #endif
 #ifdef __i386__
       case MMX1: case MMX2: case MMX3:
