@@ -8,7 +8,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -29,8 +29,8 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
-#include <stddef.h>
+#include "core.h"
+#include "memcheck/memcheck.h"
 
 /* #define DEBUG_TRANSTAB */
 
@@ -59,23 +59,10 @@ static /* const */ Int vg_tc_sector_szB = 0;
 
 /*------------------ TYPES ------------------*/
 
-#define CODE_ALIGNMENT	4	/* alignment of TCEntries */
+#define CODE_ALIGNMENT	sizeof(void*)     // alignment of TCEntries
 #define CODE_ALIGN(a)	(((a)+CODE_ALIGNMENT-1) & ~(CODE_ALIGNMENT-1))
 #define IS_ALIGNED(a)	(((a) & (CODE_ALIGNMENT-1)) == 0)
 
-/* An entry in TC.  Payload always is always padded out to a 4-aligned
-   quantity so that these structs are always word-aligned. */
-typedef
-   struct { 
-      /* +0 */ Addr   orig_addr;
-      /* +4 */ UShort orig_size;
-      /* +6 */ UShort trans_size;
-      /* +8 */ UShort jump_sites[VG_MAX_JUMPS];
-      /* +16 */ Addr chain_back;
-      /* +VG_CODE_OFFSET */ UChar  payload[0];
-   }
-   TCEntry;
-
 /* An entry in TT. */
 typedef
    struct {
@@ -188,8 +175,11 @@ void unlink_jumpsite(Addr a)
    Addr target = VG_(get_jmp_dest)(a);
    Addr *pp, p;
 
-   if (target == 0)
+   if (target == 0) {
+      if (*(UInt *)a != 0x4e800021)
+	 VG_(printf)("unlink_jumpsite didn't find a jump at %x\n", a);
       return;
+   }
    a += VG_PATCHME_JMPSZ;
    for (pp = (Addr *)(target - 4); (p = *pp) != 0; pp = (Addr *)p) {
       vg_assert(p != (Addr)pp);
@@ -207,6 +197,8 @@ void unchain_refs_to_tce(TCEntry *tce)
    Addr *p, a;
 
    for (p = &tce->chain_back; (a = *p) != 0; p = (Addr *)a) {
+      if ((*(UInt *)(a - VG_PATCHME_JMPSZ) & 0xfc000003) != 0x48000000)
+	 VG_(printf)("unchain_refs_to_tce: bad p=%p a=%x tce=%p\n", p, a, tce);
       *p = 0;
       VG_(unchain_jumpsite)(a - VG_PATCHME_JMPSZ);
    }
@@ -384,6 +376,7 @@ void discard_oldest_sector ( void )
                         s, vg_tc_used[s]);
 
 #if 0
+      VALGRIND_DISCARD_TRANSLATIONS(vg_tc[s], vg_tc_used[s]);
       for(i = 0; i < VG_TC_N_SECTORS; i++) {
 	 if (i != s && vg_tc[i] != NULL)
 	    unchain_sector(i, (Addr)vg_tc[s], vg_tc_used[s]);
@@ -391,7 +384,6 @@ void discard_oldest_sector ( void )
 #else
       unchain_refs_to_sector(s);
 #endif
-
       pp_tt_tc_status ( msg );
       overall_out_count += vg_tc_stats_count[s];
       overall_out_osize += vg_tc_stats_osize[s];
@@ -427,27 +419,29 @@ Int maybe_commission_sector ( void )
    }
    for (s = 0; s < VG_TC_N_SECTORS; s++) {
       if (vg_tc[s] == NULL) {
-#if 1
-         vg_tc[s] = VG_(get_memory_from_mmap) 
-                       ( vg_tc_sector_szB, "trans-cache(sector)" );
-#else
-         // Alternative: put translations in an mmap'd file.  The main
-         // reason is to help OProfile -- OProfile can assign time spent in
-         // translations to a particular file.  The file format doesn't
-         // really matter, which is good because it's not really readable,
-         // being generated code but not a proper ELF file.
-	 Char buf[20];
-	 static Int count = 0;
-	 Int fd;
+	 if (1)
+	    vg_tc[s] = VG_(get_memory_from_mmap) 
+	       ( vg_tc_sector_szB, "trans-cache(sector)" );
+	 else {
+	    // Alternative: put translations in an mmap'd file.  The main
+	    // reason is to help OProfile -- OProfile can assign time spent in
+	    // translations to a particular file.  The file format doesn't
+	    // really matter, which is good because it's not really readable,
+	    // being generated code but not a proper ELF file.
+	    Char buf[20];
+	    static Int count = 0;
+	    Int fd;
 	 
-	 VG_(sprintf)(buf, ".transtab.%d", count++);
+	    VG_(sprintf)(buf, ".transtab.%d", count++);
 
-	 fd = VG_(open)(buf, VKI_O_RDWR|VKI_O_CREAT|VKI_O_TRUNC, 0700);
-	 //VG_(unlink)(buf);
-	 VG_(do_syscall)(__NR_ftruncate, fd, PGROUNDUP(vg_tc_sector_szB));
-	 vg_tc[s] = VG_(mmap)(0, PGROUNDUP(vg_tc_sector_szB), VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC, VKI_MAP_SHARED, 0, fd, 0);
-	 VG_(close)(fd);
-#endif
+	    fd = VG_(open)(buf, VKI_O_RDWR|VKI_O_CREAT|VKI_O_TRUNC, 0700);
+	    //VG_(unlink)(buf);
+	    VG_(do_syscall)(__NR_ftruncate, fd, PGROUNDUP(vg_tc_sector_szB));
+	    vg_tc[s] = VG_(mmap)(0, PGROUNDUP(vg_tc_sector_szB), VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC, VKI_MAP_SHARED, 0, fd, 0);
+	    VG_(close)(fd);
+	 }
+
+	 VALGRIND_MAKE_WRITABLE(vg_tc[s], vg_tc_sector_szB);
          vg_tc_used[s] = 0;
          VG_(sprintf)(msg, "after  allocation of sector %d (size %d)", 
                            s, vg_tc_sector_szB );
@@ -545,6 +539,30 @@ void VG_(sanity_check_tt_tc) ( void )
       /* It should agree with the TC entry on the orig_addr.  This may
          be VG_TTE_DELETED, or a real orig addr. */
       vg_assert(tte->orig_addr == tce->orig_addr);
+#ifdef __powerpc__
+      {
+	 Addr *p, a, loc, dst;
+	 Int j;
+	 dst = (Addr) &tce->payload;
+	 for (p = &tce->chain_back; (a = *p) != 0; p = (Addr *)a) {
+	    loc = a - VG_PATCHME_JMPSZ;
+	    if (*(UInt *)loc != 0x48000000 + ((dst - loc) & 0x03fffffc))
+	       VG_(printf)("back chain wrong for tce=%p a=%x\n", tce, a);
+	    vg_assert(*(UInt *)loc == 0x48000000 + ((dst - loc) & 0x03fffffc));
+	 }
+	 for (j = 0; j < VG_MAX_JUMPS; ++j) {
+	    UShort idx;
+	    idx = tce->jump_sites[j];
+	    if (idx != (UShort) -1) {
+	       p = (Addr *) (tce->payload + idx);
+	       if ((*p & 0xfc000003) != 0x48000000 && *p != 0x4e800021)
+		  VG_(printf)("bad jump_site index [%d] %x (%x %x %x)\n",
+			      j, idx, p[-1], p[0], p[1]);
+	       vg_assert((*p & 0xfc000003) == 0x48000000 || *p == 0x4e800021);
+	    }
+	 }
+      }
+#endif
    }
 }
 
@@ -607,11 +625,12 @@ Int VG_(get_bbs_translated) ( void )
 */
 void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
                              Addr trans_addr, Int trans_size,
-			     UShort jumps[VG_MAX_JUMPS])
+			     UCodeBlock *cb )
 {
    Int i, nBytes, trans_size_aligned;
    TCEntry* tce;
    unsigned int *src, *dst, trans_words;
+   UShort prev;
 
    /*
    VG_(printf)("add_to_trans_tab(%d) %x %d %x %d\n",
@@ -637,7 +656,7 @@ void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
    tce->orig_size  = (UShort)orig_size;  /* what's the point of storing this? */
    tce->trans_size = (UShort)trans_size_aligned;
    for (i = 0; i < VG_MAX_JUMPS; i++) {
-      tce->jump_sites[i] = jumps[i];
+      tce->jump_sites[i] = cb->jumps[i];
    }
    tce->chain_back = 0;
    src = (unsigned int *) trans_addr;
@@ -646,10 +665,23 @@ void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
    for (i = 0; i < trans_words; ++i)
       dst[i] = src[i];
    flush_cache(dst, trans_size_aligned);
-   
-#ifdef __i386__
-   unchain_tce(tce);
-#endif
+
+   if (VG_(clo_trace_codegen) & 1)
+      VG_(message)(Vg_DebugMsg, "add_to_trans_tab: inserting code for %p into transtab at %p\n",
+		   orig_addr, tce->payload);
+
+   /* relocate absolute jumps */
+   prev = 0;
+   for(i = 0; i < cb->nrelocs; i++) {
+      UShort off = cb->relocs[i];
+
+      vg_assert(off < trans_size);
+      vg_assert(i == 0 || off > prev);
+
+      VG_(reloc_abs_jump)(&tce->payload[off]);
+
+      prev = off;
+   }
    add_tt_entry(tce);
 
    /* Update stats. */
@@ -670,6 +702,10 @@ void VG_(add_to_trans_tab) ( Addr orig_addr,  Int orig_size,
 Addr VG_(search_transtab) ( Addr original_addr )
 {
    TTEntry* tte;
+
+   if (vg_tt == NULL)
+      return 0;
+
    VGP_PUSHCC(VgpSlowFindT);
    tte = search_tt ( original_addr );
    if (tte == NULL) {
@@ -687,6 +723,7 @@ Addr VG_(search_transtab) ( Addr original_addr )
    }
 }
 
+
 /* Invalidate translations of original code [start .. start + range - 1].
    This is slow, so you *really* don't want to call it very often.
    Set 'unchain_blocks' if the translation being invalidated may be chained
@@ -698,6 +735,10 @@ void VG_(invalidate_translations) ( Addr start, UInt range, Bool unchain_blocks
    UInt     out_count, out_osize, out_tsize;
    Int      i;
    TCEntry* tce;
+
+   if (vg_tt == NULL)
+      return;
+
 #  ifdef DEBUG_TRANSTAB
    VG_(sanity_check_tt_tc)();
 #  endif
@@ -722,7 +763,7 @@ void VG_(invalidate_translations) ( Addr start, UInt range, Bool unchain_blocks
       tce->orig_addr = VG_TTE_DELETED;
 
 #if 0
-      if (unchain_blocks) {
+      if (VG_(clo_chain_bb) && unchain_blocks) {
          /* make sure no other blocks chain to the one we just discarded */
 	 Int j;
          for(j = 0; j < VG_TC_N_SECTORS; j++) {
@@ -825,7 +866,7 @@ extern void VG_(helper_cache_inval)(Addr start)
    }
 
 #  ifdef DEBUG_TRANSTAB
-   VG_(sanity_check_tc_tt)();
+   VG_(sanity_check_tt_tc)();
 #  endif
 
    if (VG_(clo_verbosity) > 2)
@@ -848,6 +889,9 @@ void VG_(init_tt_tc) ( void )
       TCEntries. */
    vg_assert((VG_MAX_JUMPS % 2) == 0);
 
+   // Otherwise lots of things go wrong... 
+   vg_assert(VG_CODE_OFFSET == sizeof(TCEntry));
+   
    /* Figure out how big each sector should be.  */
    vg_tc_sector_szB 
       = (VG_TT_LIMIT /* max TT entries we expect */
