@@ -1101,12 +1101,12 @@ const struct SyscallTableEntry VGA_(syscall_table)[] = {
 
    GENX_(__NR_tgkill,            sys_tgkill),         // 270
    GENX_(__NR_utimes,            sys_utimes),         // 271
-   //   (__NR_fadvise64_64,      sys_fadvise64_64),   // 272 */(Linux?)
+   LINX_(__NR_fadvise64_64,      sys_fadvise64_64),   // 272 */(Linux?)
    GENX_(__NR_vserver,           sys_ni_syscall),     // 273
    //   (__NR_mbind,             sys_mbind),          // 274 ?/?
 
-   //   (__NR_get_mempolicy,     sys_get_mempolicy),  // 275 ?/?
-   //   (__NR_set_mempolicy,     sys_set_mempolicy),  // 276 ?/?
+   LINXY(__NR_get_mempolicy,     sys_get_mempolicy),  // 275 ?/?
+   LINX_(__NR_set_mempolicy,     sys_set_mempolicy),  // 276 ?/?
    GENXY(__NR_mq_open,           sys_mq_open),        // 277
    GENX_(__NR_mq_unlink,         sys_mq_unlink),      // (mq_open+1)
    GENX_(__NR_mq_timedsend,      sys_mq_timedsend),   // (mq_open+2)
@@ -1115,6 +1115,17 @@ const struct SyscallTableEntry VGA_(syscall_table)[] = {
    GENX_(__NR_mq_notify,         sys_mq_notify),      // (mq_open+4)
    GENXY(__NR_mq_getsetattr,     sys_mq_getsetattr),  // (mq_open+5)
    GENX_(__NR_sys_kexec_load,    sys_ni_syscall),     // 283
+   // GENXY(__NR_waitid,            sys_waitid),         // 284
+   GENX_( 285,                    sys_ni_syscall ),     // 285
+//   LINX_(__NR_add_key,           sys_add_key),        // 286
+//   LINX_(__NR_request_key,       sys_request_key),    // 287
+//   LINXY(__NR_keyctl,            sys_keyctl),         // 288
+//   LINX_(__NR_ioprio_set,        sys_ioprio_set),     // 289
+
+//   LINX_(__NR_ioprio_get,        sys_ioprio_get),     // 290
+   LINX_(__NR_inotify_init,      sys_inotify_init),   // 291
+   LINX_(__NR_inotify_add_watch, sys_inotify_add_watch), // 292
+   LINX_(__NR_inotify_rm_watch,  sys_inotify_rm_watch), // 293
 };
 
 const UInt VGA_(syscall_table_size) = 
