@@ -61,7 +61,7 @@
 #include <stdarg.h>
 
 #undef __ppc__
-#define __ppc__   1	// Architecture we're installed on
+#define __ppc__   1	/* Architecture we're installed on */
 
 
 /* If we're not compiling for our target architecture, don't generate
@@ -119,7 +119,7 @@
     _zzq_args[2] = (unsigned int)(_zzq_arg2);			\
     _zzq_args[3] = (unsigned int)(_zzq_arg3);			\
     _zzq_args[4] = (unsigned int)(_zzq_arg4);			\
-    asm volatile("roll $29, %%eax ; roll $3, %%eax\n\t"		\
+    __asm__ volatile("roll $29, %%eax ; roll $3, %%eax\n\t"	\
                  "rorl $27, %%eax ; rorl $5, %%eax\n\t"		\
                  "roll $13, %%eax ; roll $19, %%eax"		\
                  : "=d" (_zzq_rlval)				\
@@ -127,7 +127,7 @@
                  : "cc", "memory"				\
                 );						\
   }
-#endif   // __x86__
+#endif   /* __x86__ */
 
 #ifdef __powerpc__
 #define VALGRIND_MAGIC_SEQUENCE(					\
@@ -149,9 +149,9 @@
 		 : "memory");						\
     _zzq_rlval = (__typeof__(_zzq_rlval)) _zzq_tmp;			\
   }
-#endif   // __powerpc__
+#endif   /* __powerpc__ */
 
-// Insert assembly code for other architectures here...
+/* Insert assembly code for other architectures here... */
 
 #else  /* NVALGRIND */
 /* Define NVALGRIND to completely remove the Valgrind magic sequence
@@ -206,7 +206,12 @@ typedef
 
           /* Allow printfs to valgrind log. */
           VG_USERREQ__PRINTF = 0x1401,
-          VG_USERREQ__PRINTF_BACKTRACE = 0x1402
+          VG_USERREQ__PRINTF_BACKTRACE = 0x1402,
+
+          /* Stack support. */
+          VG_USERREQ__STACK_REGISTER   = 0x1501,
+          VG_USERREQ__STACK_DEREGISTER = 0x1502,
+          VG_USERREQ__STACK_CHANGE     = 0x1503,
    } Vg_ClientRequest;
 
 #ifndef __GNUC__
@@ -394,4 +399,30 @@ VALGRIND_PRINTF_BACKTRACE(const char *format, ...)
                             pool, addr, 0, 0);                     \
    }
 
+/* Mark a piece of memory as being a stack. Returns a stack id. */
+#define VALGRIND_STACK_REGISTER(start, end)                        \
+   ({unsigned int _qzz_res;                                        \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__STACK_REGISTER,            \
+                            start, end, 0, 0);                     \
+    _qzz_res;                                                      \
+   })
+
+/* Unmark the piece of memory associated with a stack id as being a
+   stack. */
+#define VALGRIND_STACK_DEREGISTER(id)                              \
+   {unsigned int _qzz_res;                                         \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__STACK_DEREGISTER,          \
+                            id, 0, 0, 0);                          \
+   }
+
+/* Change the start and end address of the stack id. */
+#define VALGRIND_STACK_CHANGE(id, start, end)                      \
+   {unsigned int _qzz_res;                                         \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__STACK_CHANGE,              \
+                            id, start, end, 0);                    \
+   }
+
 #endif   /* __VALGRIND_H */
