@@ -1085,6 +1085,13 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
 	    uInstr3(cb, TAG2, 0, TempReg, td, TempReg, qd, Lit16, Tag_CLZ4_TQ);
 	    break;
 	 }
+
+         case RDTB:
+	    /* The result is completely defined. */
+	    sk_assert(u_in->tag2 == TempReg);
+	    uInstr1(cb, SETV, 4, TempReg, SHADOW(u_in->val2));
+	    VG_(copy_UInstr)(cb, u_in);
+	    break;
 #endif
 
          /* xor#(qs,qd) = qs `UifU` qd */
